// Copyright 2023 Maciej Sobczak
// This code is an idea demonstrator, not a final "product" and
// you are encouraged to improve it according to your actual needs.
// This software is free, like in the Boost Software License:
// https://opensource.org/licenses/BSL-1.0

#include "stream_db.h"
#include <iostream>
#include <map>
#include <string>
#include <vector>

// full name -> mail,phone
std::map<std::string, std::pair<std::string, std::string>> data_;

class phonebook_db : public stream_db::store
{
public:
    phonebook_db()
        : total_records(0), remove_and_modify_records(0)
    {
    }
    
    stream_db::result open(const std::string & file_name)
    {
        stream_db::result res = store::open(file_name);
        if (res == stream_db::ok)
        {
            res = replay();
        }
        
        if (res == stream_db::ok)
        {
            std::cout << "data loaded with " << total_records << " database events\n";
            
            // compact the data file if there are too many mofication records
            
            double fluff_factor = static_cast<double>(remove_and_modify_records) / total_records;
            if (fluff_factor >= 0.25)
            {
                std::cout << "the data file deserves some cleanup...\n";
                
                close();
                
                std::string archive_file_name = file_name + ".old";
                int r = std::rename(file_name.c_str(), archive_file_name.c_str());
                if (r == 0)
                {
                    std::cout << "    data file was archived as " << archive_file_name << '\n';
                }
                else
                {
                    res = stream_db::io_error;
                }
                
                if (res == stream_db::ok)
                {
                    res = store::open(file_name, stream_db::write_from_scratch);
                }
                
                if (res == stream_db::ok)
                {
                    res = dump_data();
                }
                
                if (res == stream_db::ok)
                {
                    res = sync();
                }
                
                if (res == stream_db::ok)
                {
                    std::cout << "    data file was recreated (compacted)\n";
                }
                else
                {
                    std::cout << "    error while recreating data file, new file is corrupted\n";
                }
            }
        }
        
        return res;
    }
    
    stream_db::result dump_data()
    {
        for (const auto & entry : data_)
        {
            const std::string & full_name = entry.first;
            const std::string & mail = entry.second.first;
            const std::string & phone = entry.second.second;
            
            std::vector<std::string> words;
            words.reserve(3);
            words.push_back(full_name);
            words.push_back(mail);
            words.push_back(phone);
            
            stream_db::result res = append_record("add", words);
            if (res != stream_db::ok)
            {
                return res;
            }
        }
        
        return stream_db::ok;
    }
    
private:
    virtual stream_db::result replay_add(const std::vector<std::string> & words) override
    {
        // the record has 3 fields:
        // - full name
        // - mail address
        // - phone number
        
        if (words.size() != 3)
        {
            std::cerr << "corrupted data file\n";
            return stream_db::parse_error;
        }
        
        const std::string & full_name = words[0];
        const std::string & mail = words[1];
        const std::string & phone = words[2];
        
        data_[full_name] = std::make_pair(mail, phone);
        
        ++total_records;
        
        return stream_db::ok;
    }

    virtual stream_db::result replay_remove(const std::vector<std::string> & words) override
    {
        // entries are removed by full name
        
        if (words.size() != 1)
        {
            std::cerr << "corrupted data file\n";
            return stream_db::parse_error;
        }
        
        const std::string & full_name = words[0];
        
        data_.erase(full_name);
        
        ++total_records;
        ++remove_and_modify_records;
        
        return stream_db::ok;
    }

    virtual stream_db::result replay_modify(const std::vector<std::string> & words) override
    {
        // modify record has full name, the indication of what is changed (mail/phone) and new value
        
        if (words.size() != 3)
        {
            std::cerr << "corrupted data file\n";
            return stream_db::parse_error;
        }
        
        const std::string & full_name = words[0];
        const std::string & field = words[1];
        if (field == "mail")
        {
            const std::string & mail = words[2];
            
            data_[full_name].first = mail;
        }
        else if (field == "phone")
        {
            const std::string & phone = words[2];
            
            data_[full_name].second = phone;
        }
        else
        {
            std::cerr << "corrupted data file\n";
            return stream_db::parse_error;
        }
        
        ++total_records;
        ++remove_and_modify_records;
        
        return stream_db::ok;
    }
    
    int total_records;
    int remove_and_modify_records;
};

// helper for reading line,
// takes into account a possible \n in the buffer
std::string read_line()
{
    std::string line;
    int count = 1;
    while (line.empty() && count <= 2)
    {
        std::getline(std::cin, line);
        ++count;
    }
    
    return line;
}

// actual application code
void run(phonebook_db & db)
{
    bool keep_running = true;
    while (keep_running)
    {
        std::cout << '\n';
        std::cout << "1. add phonebook entry\n";
        std::cout << "2. find\n";
        std::cout << "3. update\n";
        std::cout << "4. remove\n";
        std::cout << "0. exit\n";
        std::cout << "What would you like to do today? " << std::flush;
        
        int choice;
        std::cin >> choice;
        switch (choice)
        {
        case 0:
            {
                std::cout << "Bye!\n";
                keep_running = false;
            }
            break;
        
        case 1:
            {
                std::string full_name;
                std::string mail;
                std::string phone;
                
                std::cout << "Enter full name: " << std::flush;
                full_name = read_line();
                if (full_name.empty())
                {
                    continue;
                }
                
                std::cout << "Enter mail: " << std::flush;
                mail = read_line();
                if (mail.empty())
                {
                    continue;
                }

                std::cout << "Enter phone: " << std::flush;
                phone = read_line();
                if (phone.empty())
                {
                    continue;
                }

                std::vector<std::string> words;
                words.push_back(full_name);
                words.push_back(mail);
                words.push_back(phone);
                stream_db::result res = db.append_record("add", words);
                if (res != stream_db::ok)
                {
                    std::cout << "Write error, file is corrupted\n";
                }
                
                data_[full_name] = std::make_pair(mail, phone);
            }
            
            break;
        
        case 2:
            {
                std::string part_of_full_name;

                std::cout << "Enter name or any part of it: " << std::flush;
                part_of_full_name = read_line();
                if (part_of_full_name.empty())
                {
                    continue;
                }
                
                int count = 0;
                
                for (const auto & entry : data_)
                {
                    const std::string & full_name = entry.first;
                    
                    if (full_name.find(part_of_full_name) != std::string::npos)
                    {
                        const std::string & mail = entry.second.first;
                        const std::string & phone = entry.second.second;
                        
                        std::cout << '\n';
                        std::cout << "Full name: " << full_name << '\n';
                        std::cout << "     mail: " << mail << '\n';
                        std::cout << "    phone: " << phone << '\n';
                        
                        ++count;
                    }
                }
                
                std::cout << "\nFound " << count << " entries.\n";
            }
            
            break;
            
        case 3:
            {
                std::string part_of_full_name;

                std::cout << "Enter name or any part of it: " << std::flush;
                part_of_full_name = read_line();
                if (part_of_full_name.empty())
                {
                    continue;
                }
                
                int count = 0;
                
                std::cout << "1. update mail\n";
                std::cout << "2. update phone\n";
                std::cout << "Which attribute do you want to change? " << std::flush;
                
                int attr;
                std::cin >> attr;
                if (attr == 1)
                {
                    std::string mail;
                    std::cout << "Enter mail: " << std::flush;
                    mail = read_line();
                    if (mail.empty())
                    {
                        continue;
                    }
                    
                    for (auto & entry : data_)
                    {
                        const std::string & full_name = entry.first;
                        
                        if (full_name.find(part_of_full_name) != std::string::npos)
                        {
                            std::vector<std::string> words;
                            words.push_back(full_name);
                            words.push_back("mail");
                            words.push_back(mail);
                            stream_db::result res = db.append_record("modify", words);
                            if (res != stream_db::ok)
                            {
                                std::cout << "Write error, file is corrupted\n";
                            }
                            
                            entry.second.first = mail;
                            
                            ++count;
                        }
                    }
                }
                else if (attr == 2)
                {
                    std::string phone;
                    std::cout << "Enter phone: " << std::flush;
                    phone = read_line();
                    if (phone.empty())
                    {
                        continue;
                    }
                    
                    for (auto & entry : data_)
                    {
                        const std::string & full_name = entry.first;
                        
                        if (full_name.find(part_of_full_name) != std::string::npos)
                        {
                            std::vector<std::string> words;
                            words.push_back(full_name);
                            words.push_back("phone");
                            words.push_back(phone);
                            stream_db::result res = db.append_record("modify", words);
                            if (res != stream_db::ok)
                            {
                                std::cout << "Write error, file is corrupted\n";
                            }
                            
                            entry.second.second = phone;
                            
                            ++count;
                        }
                    }
                }
                
                std::cout << "Updated " << count << " entries.\n";
            }
            
            break;
        
        case 4:
            {
                std::string part_of_full_name;

                std::cout << "Enter name or any part of it: " << std::flush;
                part_of_full_name = read_line();
                if (part_of_full_name.empty())
                {
                    continue;
                }

                std::vector<std::string> full_names_to_remove;
                
                for (const auto & entry : data_)
                {
                    const std::string & full_name = entry.first;
                    
                    if (full_name.find(part_of_full_name) != std::string::npos)
                    {
                        full_names_to_remove.push_back(full_name);
                    }
                }
                
                for (const auto & full_name : full_names_to_remove)
                {
                    data_.erase(full_name);
                    
                    std::vector<std::string> words;
                    words.push_back(full_name);
                    
                    stream_db::result res = db.append_record("remove", words);
                    if (res != stream_db::ok)
                    {
                        std::cout << "Error writing to data file.\n";
                    }
                }
                
                std::cout << "Removed " << full_names_to_remove.size() << " entries.\n";
            }
            
            break;
        }
    }
}

int main(int argc, char * argv[])
{
    if (argc != 2)
    {
        std::cout << "need data file name\n";
        return EXIT_FAILURE;
    }
    
    std::string file_name(argv[1]);
    
    phonebook_db db;
    
    stream_db::result res = db.open(file_name);
    if (res == stream_db::ok)
    {
        run(db);
    }
    else
    {
        std::cout << "cannot open file " << file_name << '\n';
    }
}
