// Copyright 2023 Maciej Sobczak
// This code is an idea demonstrator, not a final "product" and
// you are encouraged to improve it according to your actual needs.
// This software is free, like in the Boost Software License:
// https://opensource.org/licenses/BSL-1.0

#include "stream_db.h"

using namespace stream_db;

store::store()
    : file_(NULL), mode_(read_only)
{
}

store::~store()
{
    close();
}

result store::open(const std::string & file_name, access_mode mode)
{
    result res;

    mode_ = mode;
    const char * fopen_mode;
    switch (mode)
    {
    case read_only:
        fopen_mode = "r";
        break;
    case read_write:
        fopen_mode = "r+";
        break;
    case write_from_scratch:
    default:
        fopen_mode = "w";
        break;
    }
        
    file_ = std::fopen(file_name.c_str(), fopen_mode);
    if (file_ != NULL)
    {
        res = ok;
    }
    else
    {
        res = io_error;
    }
    
    return res;
}

result store::replay()
{
    result res;
    
    if ((file_ != NULL) &&
        ((mode_ == read_only) || (mode_ == read_write)))
    {
        enum parse_state { command, whitespace, word, length_prefix, sized_string };

        std::string cmd;
        std::string w;
        int length;
        std::vector<std::string> words;
        
        parse_state state = command;
        
        res = ok;
        while (res == ok)
        {
            int ic = std::fgetc(file_);
            if (ic == EOF)
            {
                break;
            }
            
            char c = static_cast<char>(ic);
            
            switch (state)
            {
            case command:
                if (std::isalpha(ic) != 0)
                {
                    cmd += c;
                }
                else if (c == '\n')
                {
                    if (cmd.empty() == false)
                    {
                        res = process_record(cmd, words);
                        cmd.clear();
                        words.clear();
                    }
                    
                    // stay in command
                }
                else if (std::isspace(ic) != 0)
                {
                    state = whitespace;
                }
                else
                {
                    res = parse_error;
                }
                
                break;
            
            case whitespace:
                if (c == '\n')
                {
                    res = process_record(cmd, words);
                    cmd.clear();
                    words.clear();
                    
                    state = command;
                }
                else if (std::isspace(ic) != 0)
                {
                    // stay in whitespace
                }
                else if (c == '[')
                {
                    length = 0;
                    
                    state = length_prefix;
                }
                else
                {
                    w += c;
                    
                    state = word;
                }
                
                break;
            
            case word:
                if (c == '\n')
                {
                    words.push_back(w);
                    w.clear();
                    
                    res = process_record(cmd, words);
                    cmd.clear();
                    words.clear();
                    
                    state = command;
                }
                else if (std::isspace(ic) != 0)
                {
                    words.push_back(w);
                    w.clear();
                    
                    state = whitespace;
                }
                else
                {
                    w += c;
                    
                    // stay in word
                }
                
                break;
            
            case length_prefix:
                if (std::isdigit(ic) != 0)
                {
                    int digit = c - '0';
                    
                    length *= 10;
                    length += digit;
                }
                else if (c == ']')
                {
                    state = sized_string;
                }
                else
                {
                    res = parse_error;
                }
                
                break;
            
            case sized_string:
            
                w += c;
                --length;
                if (length == 0)
                {
                    words.push_back(w);
                    w.clear();
                    
                    state = whitespace;
                }
                
                break;
            }
        }
    }
    else
    {
        res = bad_state;
    }
    
    return res;
}

result store::append_record(
    const std::string & cmd, const std::vector<std::string> & words)
{
    result res;
    
    if ((file_ != NULL) &&
        ((mode_ == read_write) || (mode_ == write_from_scratch)))
    {
        res = ok;
        
        std::size_t written = std::fwrite(cmd.data(), 1, cmd.size(), file_);
        if (written != cmd.size())
        {
            res = io_error;
        }
        
        auto it = words.begin();
        auto end = words.end();
        for ( ; (res == ok) && (it != end); ++it)
        {
            const std::string & w = *it;
            std::size_t wsize = w.size();
            
            bool needs_wrapping = false;
            if (wsize == 0)
            {
                needs_wrapping = true;
            }
            else if (w[0] == '[')
            {
                needs_wrapping = true;
            }
            else
            {
                auto wit = w.begin();
                auto wend = w.end();
                for ( ; (needs_wrapping == false) && (wit != wend); ++wit)
                {
                    if (std::isspace(*wit) != 0)
                    {
                        needs_wrapping = true;
                    }
                }
            }
            
            int r = std::fputc(' ', file_);
            if (r == EOF)
            {
                res = io_error;
            }
            
            if ((res == ok) && needs_wrapping)
            {
                int printed = std::fprintf(file_, "[%d]", static_cast<int>(wsize));
                if (printed <= 0)
                {
                    res = io_error;
                }
            }
            
            if ((res == ok) && (wsize > 0))
            {
                written = std::fwrite(w.data(), 1, wsize, file_);
                if (written != wsize)
                {
                    res = io_error;
                }
            }
        }

        if (res == ok)
        {
            int r = std::fputc('\n', file_);
            if (r == EOF)
            {
                res = io_error;
            }
        }
    }
    else
    {
        res = bad_state;
    }
    
    return res;
}

result store::process_record(
    const std::string & cmd, const std::vector<std::string> & words)
{
    result res = ok;
    
    // capture exceptions from user callbacks
    try
    {
        if (cmd == "add")
        {
            res = replay_add(words);
        }
        else if (cmd == "modify")
        {
            res = replay_modify(words);
        }
        else if (cmd == "remove")
        {
            res = replay_remove(words);
        }
        else
        {
            // unknown command in file
            
            res = replay_custom_command(cmd, words);
        }
    }
    catch (...)
    {
        res = callback_error;
    }
    
    return res;
}

result store::sync()
{
    result res;
    
    if (file_ != NULL)
    {
        int r = std::fflush(file_);
        if (r == 0)
        {
            res = ok;
        }
        else
        {
            res = io_error;
        }
    }
    else
    {
        res = bad_state;
    }
    
    return res;
}

void store::close()
{
    std::fclose(file_);
    file_ = NULL;
}

result store::replay_add(const std::vector<std::string> & /* words */)
{
    return parse_error;
}

result store::replay_remove(const std::vector<std::string> & /* words */)
{
    return parse_error;
}

result store::replay_modify(const std::vector<std::string> & /* words */)
{
    return parse_error;
}

result store::replay_custom_command(
    const std::string & /* cmd */, const std::vector<std::string> & /* words */)
{
    return parse_error;
}
